"""Reads and writes XML for Axiell EMu"""
from .containers import EMuColumn, EMuConfig, EMuGrid, EMuRow, EMuRecord, EMuSchema
from .io import EMuReader, write_import
from .types import EMuDate, EMuFloat, EMuType
from .utils import (
    get_mod,
    has_mod,
    is_nesttab,
    is_nesttab_inner,
    is_ref,
    is_ref_tab,
    is_tab,
    strip_mod,
    strip_tab,
)

__version__ = "0.1b1"
__author__ = "Adam Mansur"
__credits__ = "Smithsonian National Museum of Natural History"



"""Defines objects used to read and write XML for Axiell EMu"""
import datetime as dt
import glob
import json
import logging
import os
import zipfile

from lxml import etree


logger = logging.getLogger(__name__)


class EMuReader:
    """Read records from an EMu XML file into dicts

    Parameters
    ----------
    path : str or Path
        path to a file or directory
    json_path : str or Path
        path to a JSON file used to cache records for faster reading

    Attributes
    ----------
    path : str or Path
        path to a file or directory
    json_path : str or Path
        path to a JSON file used to cache records for faster reading
    files : list
        list of file-like objects, each of which is an EMu XML file
    module : str
        the name of an EMu module
    """

    #: EMuConfig : module-wide configuration parameters. Set automatically
    #: when an EMuConfig object is created.
    config = None

    #: EMuSchema : info about a specific EMu configuration. Set automatically
    #: when an EMuSchema object is created.
    schema = None

    def __init__(self, path, json_path=None):
        self.path = path
        self._rec_class = dict
        self.json_path = json_path
        self.files = []
        self.module = None
        self._get_files()
        self._load_schema()

    def __iter__(self):
        for rec in self.from_file():
            yield rec

    def from_file(self):
        """Reads data from file, using JSON if possible

        Yields
        ------
        dict
            EMu record
        """
        if not self.json_path:
            return self.from_xml()

        # If JSON is older than the newest XML file, regenerate it
        try:
            if os.path.getmtime(self.json_path) < self.files[-1].getmtime():
                logger.info("Regenerating JSON (XML is newer)")
                self.to_json()
        except FileNotFoundError:
            logger.info("Generating JSON (JSON not found)")
            self.to_json()

        return self.from_json()

    def from_xml(self):
        """Reads data from XML

        Yields
        ------
        dict
            EMu record
        """
        for filelike in self.files:
            logger.info("Reading records from %s", filelike)
            i = 0
            with filelike.open("rb") as source:
                try:
                    context = etree.iterparse(source, events=["end"], tag="tuple")
                    for _, element in context:
                        # Process children of module table only
                        parent = element.getparent().get("name")
                        if parent is not None and parent.startswith("e"):
                            try:
                                yield self._parse(element)
                            finally:
                                element.clear()
                                # while element.getprevious() is not None:
                                #    del element.getparent()[0]
                                i += 1
                                if not i % 5000:
                                    logger.info("Read %s records from %s", i, filelike)
                finally:
                    del context
            logger.info("Read %s records total", i)

    def from_json(self, chunk_size=2097152):
        """Reads data from JSON

        Parameters
        ----------
        chunk_size : int
            size of chunk to use when reading the file

        Yields
        ------
        dict
            EMu record
        """
        logger.info("Reading records from %s", self.json_path)
        i = 0
        with open(self.json_path, encoding="utf-8") as f:
            f.read(1)
            add_to_next_chunk = []
            while True:
                chunk = f.read(chunk_size)
                if add_to_next_chunk:
                    chunk = "".join(add_to_next_chunk[::-1]).lstrip(",") + chunk
                    add_to_next_chunk = []

                if len(chunk) <= 1:
                    break

                while True:
                    try:
                        for rec in json.loads(f"[{chunk.lstrip(',')[:-1]}]"):
                            try:
                                yield rec
                            finally:
                                i += 1
                                if not i % 5000:
                                    logger.info(
                                        "Read %s records from %s", i, self.json_path
                                    )
                        break
                    except json.JSONDecodeError:
                        chunk, trailer = chunk.rsplit("{", 1)
                        add_to_next_chunk.append(f"{{{trailer}")
        logger.info("Read %s records total", i)

    def to_json(self, path=None, **kwargs):
        """Writes JSON version of XML to file

        Parameters
        ----------
        path : str
            path to write JSON
        kwargs :
            keyword arguments for json.dump()
        """
        if path is None:
            path = self.json_path

        logger.info("Writing records from %s to JSON", self.path)

        params = {
            "ensure_ascii": False,
            "indent": None,
            "sort_keys": False,
            "separators": (",", ":"),
        }
        params.update(**kwargs)

        sep = params["separators"][0]

        with open(path, "w", encoding="utf-8"):
            pass

        try:
            with open(path, "a", encoding="utf-8") as f:
                f.write("[")
                records = []
                for rec in self.from_xml():
                    records.append(rec)
                    if len(records) > 1000:
                        f.write(json.dumps(records, **params)[1:-1] + sep)
                        records = []
                if records:
                    f.write(json.dumps(records, **params)[1:-1])
                f.write("]")
        except KeyboardInterrupt as exc:
            # Remove the partial JSON file if write is interrupted
            os.remove(path)
            raise IOError("Conversion to JSON failed") from exc

    def _parse(self, xml):
        """Parses a record from XML

        Parameters
        ----------
        xml : lxml.Element
            XML representing a single record

        Returns
        -------
        dict
           EMu record
        """
        if self._rec_class != dict:
            dct = self._rec_class(module=self.module)
        else:
            dct = self._rec_class()

        elements = [(dct, "", xml)]
        while elements:
            new_elems = []
            for obj, parent_name, elem in elements:
                for child in elem:

                    # Get field name
                    name = child.get("name")
                    if name is None:
                        name = ""

                    # Get field text
                    text = child.text
                    if text is not None:
                        text = text.strip()

                    # Add an atomic field with a value
                    if child.tag == "atom" and (text or not parent_name):
                        try:
                            obj[name] = text
                        except TypeError:
                            obj.append(text)

                    # Add a reference
                    elif child.tag == "tuple" and name.endswith("Ref"):
                        obj[name] = {}
                        new_elems.append((obj[name], name, child))

                    # Add a table or reference table
                    elif child.tag == "table" or (child.tag == "tuple" and name):
                        try:
                            obj[name] = []
                            new_elems.append((obj[name], name, child))
                        except TypeError:
                            obj.append([])
                            new_elems.append((obj[-1], name, child))

                    # Add a row to a table
                    elif child.tag == "tuple" and parent_name.endswith("Ref_tab"):
                        obj.append({})
                        new_elems.append((obj[-1], name, child))

                    elif child.tag == "tuple":
                        new_elems.append((obj, name, child))

                elements = new_elems

        return dct

    def _get_files(self):
        """Analyzes source files on self.path"""
        files = []
        zip_file = None
        if self.path:
            if os.path.isdir(self.path):
                files = glob.glob(os.path.join(self.path, "*.xml"))
            elif self.path.lower().endswith(".xml"):
                files = [self.path]
            elif self.path.lower().endswith(".zip"):
                zip_file = zipfile.ZipFile(self.path)
                files = zipfile.ZipFile(self.path).infolist()
            else:
                raise IOError(f"Invalid path: {self.path}")

        # Order source files from oldest to newest
        self.files = [FileLike(obj, zip_file=zip_file) for obj in files]
        self.files.sort(key=lambda f: f.getmtime())

        # Get the module name from the first table tag
        with self.files[0].open(encoding="utf-8") as f:
            for line in f:
                if line.strip().startswith("<table"):
                    self.module = line.split("=", 1)[-1].strip('">\r\n')
                    break

    def _load_schema(self):
        """Tries to load the schema based on the rec_class"""
        if self.schema is None:
            try:
                schema = self._rec_class.schema
                if schema is None:
                    # This will also load the configuration
                    schema = self._rec_class(module=self.module).schema
            except (AttributeError, ValueError):
                schema = None
        return self.schema


class FileLike:
    """Open text and zip files using the same interface

    Parameters
    ----------
    filelike : mixed
        either the path to an XML file or a ZipInfo object
    zip_file : zipfile.ZipFile
        if filelike is a ZipInfo object, the zip file containing that object

    Attributes
    ----------
    path : str
        path to file
    zip_info : zipfile.ZipInfo
        member of a zip archive
    zip_file : zipfile.ZipFile
        the zip file containing the ZipInfo object
    """

    def __init__(self, filelike, zip_file=None):
        self.path = None
        self.zip_info = None
        self.zip_file = None
        if zip_file:
            self.zip_info = filelike
            self.zip_file = zip_file
        else:
            self.path = os.path.realpath(filelike)

    def __str__(self):
        return f'<FileLike name="{self.filename}">'

    def __repr__(self):
        return str(self)

    @property
    def filename(self):
        """Name of the file-like object"""
        return os.path.basename(self.path) if self.path else self.zip_info.filename

    def open(self, mode="r", encoding=None):
        """Opens a file or ZipInfo object"""
        if not self.zip_info:
            return open(self.path, mode=mode, encoding=encoding)
        stream = self.zip_file.open(self.zip_info, mode.rstrip("b"))
        if encoding:
            return _ByteDecoder(stream, encoding)
        return stream

    def getmtime(self):
        """Returns last modification timestamp from a file or ZipInfo object"""
        try:
            return os.path.getmtime(self.path)
        except TypeError:
            return dt.datetime(*self.zip_info.date_time).timestamp()


class _ByteDecoder:
    """File-like context manager that encodes a binary stream from a zip file"""

    def __init__(self, stream, encoding):
        self._stream = stream
        self._encoding = encoding

    def __iter__(self):
        for line in self._stream:
            yield line.decode(self._encoding)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exception, traceback):
        if exception:
            raise exception
        self._stream.close()


def write_import(records, path):
    """Writes records to an EMu import file

    Parameters
    ----------
    records : list-like
        list of EMuRecords to be imported
    path : str
        path to write the import file
    """
    root = etree.Element("table")
    root.set("name", records[0].module)

    for rec in records:
        rec.to_xml(root)

    for i, rec in enumerate(root):
        rec.addprevious(etree.Comment(f"Row {i + 1}"))

    root.getroottree().write(
        path, pretty_print=True, xml_declaration=True, encoding="utf-8"
    )



"""Wrappers for data that can be garbled during read/write"""
import logging
import re
from calendar import monthrange
from datetime import date, datetime


logger = logging.getLogger(__name__)


class EMuType:
    """Container for data types that may be garbled during read/write

    For example, transforming a year to a date using datetime.strptime()
    imposes a month and date, which could be bad news if that data is ever
    loaded back into the database. This class tracks the original string
    and format while coercing the string to a Python data type and
    providing support for basic operations.

    Parameters
    ----------
    val : mixed
        value to wrap
    fmt : str
        formatting string used to translate value back to a string

    Attributes
    ----------
    value : mixed
        value coerced to the correct type from a string
    format : str
        a formatting string
    verbatim : mixed
        the original, unparsed value
    """

    def __init__(self, val, fmt="{}"):
        self.verbatim = val
        self.value = val
        self.format = fmt

    def __str__(self):
        return self.format.format(self.value)

    def __repr__(self):
        return f"{self.__class__.__name__}('{str(self)}')"

    def __eq__(self, other):
        if self.is_range():
            other = self.coerce(other)
            return (
                self.value == other.value
                and self.min_value == other.min_value
                and self.max_value == other.max_value
            )
        return self.value == other

    def __ne__(self, other):
        if self.is_range():
            other = self.coerce(other)
            return (
                self.value != other.value
                or self.min_value != other.min_value
                or self.max_value != other.max_value
            )
        return self.value != other

    def __lt__(self, other):
        if self.is_range():
            other = self.coerce(other)
            return self.max_value < other.min_value
        return self.value < other

    def __le__(self, other):
        if self.is_range():
            other = self.coerce(other)
            return self.min_value <= other.max_value
        return self.value <= other

    def __gt__(self, other):
        if self.is_range():
            other = self.coerce(other)
            return self.min_value > other.max_value
        return self.value > other

    def __ge__(self, other):
        if self.is_range():
            other = self.coerce(other)
            return self.max_value >= other.min_value
        return self.value >= other

    def __contains__(self, other):
        if self.is_range():
            other = self.coerce(other)
            return (
                self.min_value <= other.min_value and self.max_value >= other.max_value
            )
        raise ValueError(f"{self.__class__.__name__} is not a range")

    def __add__(self, other):
        return self._math_op(other, "__add__")

    def __sub__(self, other):
        return self._math_op(other, "__sub__")

    def __mul__(self, other):
        return self._math_op(other, "__mul__")

    def __floordiv__(self, other):
        return self._math_op(other, "__floordiv__")

    def __div__(self, other):
        return self._math_op(other, "__div__")

    def __truediv__(self, other):
        return self._math_op(other, "__truediv__")

    def __mod__(self, other):
        return self._math_op(other, "__mod__")

    def __divmod__(self, other):
        return self._math_op(other, "__divmod__")

    def __pow__(self, other):
        return self._math_op(other, "__pow__")

    def __iadd__(self, other):
        result = self + other
        self.value = result.value
        self.format = result.format
        return self

    def __isub__(self, other):
        result = self - other
        self.value = result.value
        self.format = result.format
        return self

    def __imul__(self, other):
        result = self * other
        self.value = result.value
        self.format = result.format
        return self

    def __ifloordiv__(self, other):
        result = self // other
        self.value = result.value
        self.format = result.format
        return self

    def __idiv__(self, other):
        result = self / other
        self.value = result.value
        self.format = result.format
        return self

    def __itruediv__(self, other):
        result = self / other
        self.value = result.value
        self.format = result.format
        return self

    def __imod__(self, other):
        result = self % other
        self.value = result.value
        self.format = result.format
        return self

    def __ipow__(self, other):
        result = self ** other
        self.value = result.value
        self.format = result.format
        return self

    @property
    def min_value(self):
        """Minimum value needed to express the original string"""
        return self.value

    @property
    def max_value(self):
        """Maximum value needed to express the original string"""
        return self.value

    def coerce(self, other):
        """Coerces another object to the current class

        Parameters
        ----------
        other : mixed
            an object to convert to this class

        Returns
        -------
        EMuType
            other as EMuType
        """
        if not isinstance(other, self.__class__):
            other = self.__class__(other)
        return other

    def is_range(self):
        """Checks if class represents a range"""
        return self.min_value != self.max_value

    def _math_op(self, other, operation):
        """Performs the specified arithmetic operation"""

        if self.is_range():
            min_val = self.__class__(self.min_value)._math_op(other, operation)
            max_val = self.__class__(self.max_value)._math_op(other, operation)
            return (min_val, max_val)

        if isinstance(other, self.__class__):
            val = getattr(self.value, operation)(other.value)
            # Use the more precise format for add/substract
            i = 1 if operation in {"__add__", "__sub__"} else 0
            fmt = sorted([self.format, other.format])[i]
        else:
            val = getattr(self.value, operation)(other)
            fmt = self.format

        if isinstance(val, tuple):
            return tuple([self.__class__(str(val), fmt=fmt) for val in val])

        try:
            return self.__class__(str(val), fmt=fmt)
        except ValueError:
            # Some operations return values that cannot be coerced to the original
            # class, for example, subtracting one date from another
            return val


class EMuFloat(EMuType):
    """Wraps floats read from strings to preserve precision

    Parameters
    ----------
    val : str or float
        float as a string or float
    fmt : str
        formatting string used to convert the float back to a string. Computed
        for strings but must be included if val is a float.

    Attributes
    ----------
    value : float
        float parsed from string
    format : str
        formatting string used to convert the float back to a string
    verbatim : mixed
        the original, unparsed value
    """

    def __init__(self, val, fmt=None):
        """Initialize an EMuFloat object

        Parameters
        ----------
        val : str or float
            the number to wrap
        fmt : str
            a Python formatting string. Must be probided if val is a float,
            otherwise it will be determined from val.
        """

        self.verbatim = val

        fmt_provided = fmt is not None

        if isinstance(val, float) and not fmt_provided:
            raise ValueError("Must provide fmt when passing a float")

        if isinstance(val, self.__class__):
            self.value = val.value
            self.format = val.format
            val = str(val)  # convert to string so the verification step works
        elif fmt_provided:
            self.value = float(val)
            self.format = fmt
        else:
            self.value = float(val)
            num_digits = len(val.split(".")[1]) if "." in val else 0
            self.format = f"{{:.{num_digits}f}}"

        # Verify that the parsed value is the same as the original string if
        # the format string was calculated
        if not fmt_provided and val.lstrip("0").rstrip(".") != str(self).lstrip("0"):
            raise ValueError(f"Parsing changed value ('{val}' became '{self}'")

    def __format__(self, format_spec):
        try:
            return format(str(self), format_spec)
        except ValueError:
            return format(float(self), format_spec)

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return self.value


class EMuDate(EMuType):
    """Wraps dates read from strings to preserve meaning

    Supports addition and subtraction using timedelta objects but not augmented
    assignment using += or -=.

    Parameters
    ----------
    val : str or datetime.date
        date as a string or date object
    fmt : str
        formatting string used to convert the value back to a string. If
        omitted, the class will try to determine the correct format.

    Attributes
    ----------
    value : datetime.date
        date parsed from string
    format : str
        date format string used to convert the date back to a string
    verbatim : mixed
        the original, unparsed value
    """

    directives = {
        "day": ("%d", "%-d"),
        "month": ("%B", "%b", "%m", "%-m"),
        "year": ("%Y", "%y"),
    }
    formats = {"day": "%Y-%m-%d", "month": "%b %Y", "year": "%Y"}

    def __init__(self, val, fmt=None):
        """Initialize an EMuDate object

        Parameters
        ----------
        val : str or datetime.date
            the date
        fmt : str
            a date format string
        """

        self.verbatim = val

        fmt_provided = fmt is not None

        fmts = [
            ("day", "%Y-%m-%d"),
            ("month", "%Y-%m-"),
            ("month", "%b %Y"),
            ("year", "%Y"),
        ]

        if isinstance(val, EMuDate):
            self.value = val.value
            self.kind = val.kind
            self.format = val.format
            val = str(val)  # convert to string so the verification step works
            fmt = self.format
            fmts.clear()  # clear the format dict to skip the format check

        elif isinstance(val, date):
            self.value = val
            self.kind = "day"
            self.format = "%Y-%m-%d"
            val = str(val)  # convert to string so the verification step works
            fmt = self.format
            fmts.clear()  # clear the format dict to skip the format check

        elif fmt:
            # Assess speciicity of if custom formatting string provided
            for kind, directives in self.directives.items():
                if any((d in fmt for d in directives)):
                    parsed = datetime.strptime(val, fmt)
                    self.value = date(parsed.year, parsed.month, parsed.day)
                    self.kind = kind
                    self.format = self.formats[kind]
                    fmts.clear()
                    break

        for kind, fmt in fmts:
            try:
                parsed = datetime.strptime(str(val), fmt)
                self.value = date(parsed.year, parsed.month, parsed.day)
                self.kind = kind
                self.format = self.formats[kind]
                break
            except (TypeError, ValueError):
                pass
        else:
            if fmts:
                raise ValueError(f"Could not parse date: {repr(val)}")

        # Verify that the parsed value is the same as the original string if
        # the format string was calculated
        if not fmt_provided and str(val) != self.strftime(fmt):
            raise ValueError(f"Parsing changed value ('{val}' became '{self}'")

    def __str__(self):
        return self.value.strftime(self.format)

    def strftime(self, fmt=None):
        """Formats date as a string

        Parameters
        ----------
        fmt : str
            date format string

        Returns
        -------
        str
            date as string
        """

        if fmt is None:
            fmt = self.format

        # Forbid formats that are more specific than the original string. Users
        # can force the issue by formatting the value attribute directly.
        if not self.day:
            allowed = []
            if self.year:
                allowed.extend(self.directives["year"])
            if self.month:
                allowed.extend(self.directives["month"])

            directives = re.findall(r"%[a-z]", fmt, flags=re.I)
            disallowed = set(directives) - set(allowed)
            if disallowed:
                raise ValueError(f'Invalid directives for "{str(self)}": {disallowed}')

        return self.value.strftime(fmt)

    @property
    def min_value(self):
        """Minimum date needed to express the original string

        For example, the first day of the month for a date that specifies
        only a month and year or the first day of the year for a year.
        """
        if self.kind == "day":
            return self.value
        if self.kind == "month":
            return date(self.value.year, self.value.month, 1)
        if self.kind == "year":
            return date(self.value.year, 1, 1)
        raise ValueError(f"Invalid kind: {self.kind}")

    @property
    def max_value(self):
        """Maximum date needed to express the original string

        For example, the last day of the month for a date that specifies
        only a month and year or the last day of the year for a year.
        """
        if self.kind == "day":
            return self.value
        if self.kind == "month":
            _, last_day = monthrange(self.value.year, self.value.month)
            return date(self.value.year, self.value.month, last_day)
        if self.kind == "year":
            return date(self.value.year, 12, 31)
        raise ValueError(f"Invalid kind: {self.kind}")

    @property
    def year(self):
        """Year of the parsed date"""
        return self.value.year

    @property
    def month(self):
        """Month of the parsed date"""
        return self.value.month if self.kind != "year" else None

    @property
    def day(self):
        """Day of the parsed date"""
        return self.value.day if self.kind == "day" else None
